/*
 * Copyright 2004 Ian Holsman
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define APR_WANT_STRFUNC
#include <apr_want.h>
#include <apr_strings.h>
#include <apr_time.h>
#include <apr_network_io.h>

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h" 

#include "apr_memcache.h"
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION
#include "mod_ip_count.h"
#define DEFAULT_SERVER "127.0.0.1"

module AP_MODULE_DECLARE_DATA ip_count_module;

typedef struct {
  char *memcache_servers;
  int memcache_max_requests;
  apr_short_interval_time_t memcache_max_time;
  apr_memcache_t *mc;
  /* TODO: add config settings for these */
  apr_uint32_t min_sockets;
  apr_uint32_t smax_sockets;
  apr_uint32_t max_sockets;
  apr_uint32_t ttl_sockets;
} ip_count_config_rec;

typedef struct {
  int size;
  int nelts;
  int posn;
  int lastseenposn;
  apr_time_t times[];
} user_details;

static const char *set_servers(cmd_parms *cmd, void *config, const char *arg1 )
{
    ip_count_config_rec *cfg;
    cfg = ap_get_module_config(cmd->server->module_config, &ip_count_module );
    cfg->memcache_servers = apr_pstrdup( cmd->pool, arg1);
    return NULL;
}

static const char *set_max_time(cmd_parms *cmd, void *config, const char *arg1 )
{
    ip_count_config_rec *cfg;
    cfg = ap_get_module_config(cmd->server->module_config, &ip_count_module );
    cfg->memcache_max_time = apr_time_make( atoi( arg1) ,0);

    return NULL;
}
static const char *set_max_requests(cmd_parms *cmd, void *config, const char *arg1 )
{
    ip_count_config_rec *cfg;
    cfg = ap_get_module_config(cmd->server->module_config, &ip_count_module );
    cfg->memcache_max_requests = atoi( arg1);
    return NULL;
}

static void *create_ip_count_server_config(apr_pool_t *p, server_rec *s)
{
    ip_count_config_rec *conf = apr_palloc(p, sizeof(*conf));

    conf->memcache_servers = DEFAULT_SERVER;

    conf->memcache_max_requests = 50;
    conf->memcache_max_time = apr_time_make(120,0);
    conf->min_sockets=2;
    conf->smax_sockets = 5;
    conf->max_sockets = 10;
    conf->ttl_sockets= 600;
    conf->mc = NULL;

    return conf;
}
 
static void ip_count_child_init(apr_pool_t *p, server_rec *s)
{
    char *serverstring;
    char *serverport;
    char *last;
    int nservers=0;

    apr_status_t rv;
    apr_memcache_server_t *serverrec;

    ip_count_config_rec *conf = ap_get_module_config(s->module_config,
                                                      &ip_count_module);
    /* Find all the servers in the first run to get a total count */
    serverstring = apr_pstrdup(p, conf->memcache_servers);
    serverport = apr_strtok(serverstring, " ", &last);
    while (serverport) {
        nservers++;
        serverport = apr_strtok(NULL," ", &last);
    }    

    rv = apr_memcache_create(p, nservers, 0, &conf->mc);

    if (rv != APR_SUCCESS) {
        ap_log_error(APLOG_MARK, APLOG_CRIT, rv, s,
                  "apr_memcache_create:unable to initialize memcached");
    }
    serverstring = apr_pstrdup(p, conf->memcache_servers);
    serverport = apr_strtok(serverstring, " ", &last);
    while (serverport) {
       char *server;
       char *scope;
       apr_port_t port;

       rv = apr_parse_addr_port( &server, &scope, &port, serverport, p );
       if ( rv != APR_SUCCESS) {
            ap_log_error(APLOG_MARK, APLOG_CRIT, rv, s,
                      "apr_memcache_server_create: unable to parse %s (%s)", serverport, conf->memcache_servers );
       }
       if (server == NULL ) {
           server = DEFAULT_SERVER;
       }
       
       rv = apr_memcache_server_create(p, 
                        server, 
                        port,
                        conf->min_sockets,
                        conf->smax_sockets,
                        conf->max_sockets,
                        conf->ttl_sockets,
                        &serverrec);
        if (rv != APR_SUCCESS) {
            ap_log_error(APLOG_MARK, APLOG_CRIT, rv, s,
                      "apr_memcache_server_create:%s %d", server, port);
        }
        rv = apr_memcache_add_server(conf->mc, serverrec);
        if (rv != APR_SUCCESS) {
            ap_log_error(APLOG_MARK, APLOG_CRIT, rv, s,
                      "apr_memcache_add_server:");
        }

       serverport = apr_strtok(NULL," ", &last);
    }
}
/* Checking ID */
static int ip_count_check_auth(request_rec *r)
{
    ip_count_config_rec *conf = ap_get_module_config(r->server->module_config,
                                                      &ip_count_module);
    char *reason = NULL;
    apr_status_t rv;
    user_details *det;
    int i;
    apr_time_t oldest_time;
    char datetimestr[APR_RFC822_DATE_LEN];
    apr_time_t *last_seen;
    char *key;
    apr_memcache_t *mc;
    int new=0;
    apr_size_t len;
    apr_uint32_t flags;
    char *result;

    if (r->main) {
        return DECLINED;
    }
    key = r->connection->remote_ip;

//    Use this for debugging
//    key = r->args;
    if (!key) {
        key = "none";
    }
    mc = conf->mc;
//    mc_server_add(mc, "cn-sfo1-pma-monitor.cnet.com","11211");

    rv = apr_memcache_getp( mc, r->pool, key, &result, &len, &flags );
    if (rv != APR_SUCCESS && rv != APR_NOTFOUND) {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, rv, r,
                  "Getting cache key for IP %s", key);
        return DECLINED;
    }

    oldest_time = r->request_time - conf->memcache_max_time;
    if (rv == APR_NOTFOUND) {
        det = apr_palloc( r->pool, sizeof(user_details) + sizeof(apr_time_t)* conf->memcache_max_requests );
        det->size = conf->memcache_max_requests;
        det->nelts = 0;
        det->posn = 0;
        det->lastseenposn = 0;
        new = 1;
    }
    else {
        det = (user_details*)result;
    }

    /* remove expired ones */
    last_seen = det->times;
    while  ( det->nelts > 0 && last_seen[ det->lastseenposn ] < oldest_time) {
        det->nelts--;
        det->times[ det->lastseenposn ] = -1;
        det->lastseenposn = ( det->lastseenposn + 1 ) % det->size;
    }

    /* check if we have gone too far */
    if ( det->nelts >= det->size || det->nelts >= conf->memcache_max_requests) {
        apr_rfc822_date(datetimestr, det->times[det->lastseenposn] );
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                  "access to %s failed, reason: too many requests from IP %s started at %s", r->uri, key, datetimestr);
        return HTTP_SERVICE_UNAVAILABLE;
    }
    /* now add this one */

    det->nelts++;
    det->posn = ( det->posn + 1 ) % det->size;
    det->times[det->posn ] = r->request_time;

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "Setting cache key for IP %s OK: seen %d in the last %d seconds", key, det->nelts, apr_time_sec(conf->memcache_max_time));
    /* keep the record around for a bit longer than the max time */
    rv = apr_memcache_set( mc, 
			   key, 
			   (char*)det, 
			   sizeof(user_details) + (sizeof(apr_time_t)* det->size),
			   apr_time_sec(conf->memcache_max_time)+600,
			   0);

    if (rv) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
                  "Setting cache key for IP %s failed", key);
    }
    return DECLINED;
}


static const command_rec ip_count_cmds[] =
{
    AP_INIT_TAKE1("MemCacheServers", set_servers, NULL, RSRC_CONF, 
            "a list of servers running the memcached server (host:port)"),
    AP_INIT_TAKE1("MemCacheMaxRequests", set_max_requests, NULL, RSRC_CONF, 
            "Max number of requests before failing"),
    AP_INIT_TAKE1("MemCacheMaxTime", set_max_time, NULL, RSRC_CONF,
            "Time period in which the requests have to come (seconds)"),
    {NULL}
};

static void register_hooks(apr_pool_t *p)
{

    ap_hook_child_init(ip_count_child_init, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_fixups(ip_count_check_auth, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA ip_count_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,                         /* dir config creater */
    NULL,                         /* dir merger --- default is to override */
    create_ip_count_server_config,/* server config */
    NULL,                         /* merge server config */
    ip_count_cmds,                /* command apr_table_t */
    register_hooks                /* register hooks */
};
